// Updates: 2004.07.27


import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;


/**
 * A multistyle text zone. This actual version is specifically made to work with
 * the <code>LogViewer</code> class, but can be easily updated to work with any 
 * client (using an interface). 
 * @author Michel Deriaz
 */
public class StyledTextZone extends JPanel { 
  public static final int NORMAL = 0;
  public static final int RED = 1;
  public static final int CYAN = 2;
  public static final int MAGENTA = 3;
  public static final int BLUE = 4;
  public static final int ORANGE = 5;
  public static final int GREEN = 6;
  public static final int GREENBACK = 7;
    
  private Style[] styles;
  private int defaultStyle = 0;
  private JTextPane tp;
  private DefaultStyledDocument doc;
  private boolean lineWrap = false;
  private JScrollPane scrollPane;
  private int vertScrollBarPos = -1;
  private int horScrollBarPos = -1;


  /**
   * Creates a new StyledTextZone. 
   * @param title the name that appear on the titled border (can be null)
   * @param border the size of the empty border around the text zone
   */  
  public StyledTextZone(String title, int border) {
    setLayout(new BorderLayout());
    setBorder(BorderFactory.createEmptyBorder(border, border, border, border));
    JPanel p = new JPanel(new BorderLayout());
    if (title != null) p.setBorder(BorderFactory.createTitledBorder("  " + title + "  "));
    
    /* Creation of the different styles. */
    StyleContext sc = new StyleContext();
    styles = new Style[8];
    styles[NORMAL] = sc.addStyle(null, null);
    styles[RED] = sc.addStyle(null, null);
    StyleConstants.setForeground(styles[RED], Color.RED);
    styles[CYAN] = sc.addStyle(null, null);
    StyleConstants.setForeground(styles[CYAN], Color.CYAN);
    styles[MAGENTA] = sc.addStyle(null, null);
    StyleConstants.setForeground(styles[MAGENTA], Color.MAGENTA); 
    styles[BLUE] = sc.addStyle(null, null);
    StyleConstants.setForeground(styles[BLUE], Color.BLUE); 
    styles[ORANGE] = sc.addStyle(null, null);
    StyleConstants.setForeground(styles[ORANGE], Color.ORANGE);             
    styles[GREEN] = sc.addStyle(null, null);
    StyleConstants.setForeground(styles[GREEN], new Color(0, 150, 0)); 
    styles[GREENBACK] = sc.addStyle(null, null);
    StyleConstants.setBackground(styles[GREENBACK], Color.GREEN);
    
    doc = new DefaultStyledDocument(sc);
    
    /* Creates a JTextPane that either wrap all the lines that are too long, or
       that prints a horizontal scollbar. */
    tp = new JTextPane(doc) {
      public void setSize(Dimension d) {
        int paWidth = getParent().getSize().width;
        if (d.width < paWidth) d.width = paWidth;
        super.setSize(d);
      }
      public boolean getScrollableTracksViewportWidth() {
        return lineWrap;
      }
    };
    
    scrollPane = new JScrollPane(tp);
    p.add(scrollPane, BorderLayout.CENTER);
    add(p, BorderLayout.CENTER);
  }
 
 
  /**
   * Sets the specified boolean to indicate whether or not this StyledTextZone
   * should be editable.
   * @param editable the boolean to be set
   */ 
  public void setEditable(boolean editable) {
    tp.setEditable(editable);
  }
  

  /**
   * Sets the specified boolean to indicate whether or not the lines of this 
   * StyledTextZone should be wrapped automatically.
   * @param lineWrap the boolean to be set
   */      
  public void setLineWrap(boolean lineWrap) {
    this.lineWrap = lineWrap;
    tp.updateUI();
  }
  
  
  /**
   * Returns the text contained in this StyledTextZone.
   * @return the text
   */
  public String getText() {
    try {
      return doc.getText(0, doc.getLength());
        // We do not use tp.getText() because it returns always (no matter what
        // the entry is) "\r\n" as line separator, on a windows system.
        // doc.getText() returns exactly what is in the text zone. 
        // Note that if you write manually (during execution) in the the
        // document, "\n" are used as line separators, and if you copy-paste a
        // document from another text editor (even running on Windows) then all 
        // the line separators will become "\n". Eventually, if you copy-paste a
        // text from the text zone to another text editor, all the line 
        // separators will be transformed into the way that the platform write 
        // them ("\r\n" for windows, for instance).
    }
    catch (BadLocationException e) {
      System.out.println(e);
    }
    return null;          
  }
  

  /**
   * Returns the selected text contained in this StyledTextZone.
   * @return the selected text
   */
  public String getSelectedText() {
    return tp.getSelectedText();
  }


  /**
   * Returns the selected text's start position.
   * @return the start position
   */
  public int getSelectionStart() {
    if (tp.getSelectedText() == null) return 0;
    return tp.getSelectionStart();
  }
  

  /**
   * Returns the selected text's end position.
   * @return the end position
   */  
  public int getSelectionEnd() {
    if (tp.getSelectedText() == null) return 0;    
    return tp.getSelectionEnd();
  }
  
  
  /**
   * Sets the text of this StyledTextZone to the specified text.
   * @param text the new text to be set
   */  
  public void setText(String text) {
    try {
      doc.remove(0, doc.getLength());
    }
    catch (BadLocationException e) {
      System.out.println(e);
    }      
    print(text);
  }
  
  
  /**
   * Adds the specified text to the existing one, using the default style.
   * @param text the new text to be added
   * @see #setDefaultStyle
   */  
  public void print(String text) {
    print(text, defaultStyle);
  }
  
   
  /**
   * Adds the specified text to the existing one, using the specified style.
   * @param text the new text to be added
   * @param style the style
   */     
  public void print(String text, int style) {
    try {
      doc.insertString(doc.getLength(), text, styles[style]);
    }
    catch (BadLocationException e) {
      System.out.println(e);
    }
  }
             

  /**
   * Adds the specified text (plus a line separator) to the existing one, using
   * the default style.
   * @param text the new text to be added
   * @see #setDefaultStyle
   */    
  public void println(String text) {
    print(text + "\n");
  }
  

  /**
   * Adds the specified text (plus a line separator) to the existing one, using
   * the specified style.
   * @param text the new text to be added
   * @param style the style
   */    
  public void println(String text, int style) {
    print(text + "\n", style);
  }


  /**
   * Sets the specified style for the all document.
   * @param style the style to be set
   */
  public void setStyle(int style) {
    setStyle(style, 0, doc.getLength());
  }

  
  /**
   * Sets the specified style for the document between the <code>start</code>
   * position and the <code>end</code> position. It is important to note that 
   * <code>start</code> and <code>end</code> actually point <b>between</b>
   * caracters, therefore <code>start</code> points just before the first
   * caracter, and <code>end</code> points just after the last one.
   * @param style the style to be set
   * @param start the start position
   * @param end the end position
   */  
  public void setStyle(int style, int start, int end) {
    int caretPos = tp.getCaretPosition(); // saves the position of the caret
    try {
      doc.replace(start, end-start, doc.getText(start, end-start), styles[style]);
    }
    catch (BadLocationException e) {
      System.out.println(e);
    }
    tp.setCaretPosition(caretPos); // restores the position of the caret
  }
  
  
  /**
   * Sets the default style.
   * @param defaultStyle the style
   */  
  public void setDefaultStyle(int defaultStyle) {
    this.defaultStyle = defaultStyle;
  }
  
  
  /**
   * Returns the default style
   * @return the default style
   */  
  public int getDefaultStyle() {
    return defaultStyle;
  }
  
  
  /**
   * Modifies the font of the specified style. If <code>style == -1</code> then
   * all styles will be modified. Remember that the logical fonts (in 
   * comparison to the physical ones) are thoses that are supported by any Java
   * runtime environment, i.e. Serif, SansSerif, Monospaced, Dialog, and
   * DialogInput. 
   * @param style the style to modify, or -1 to modify all styles
   * @param font the name of the font
   */  
  public void setFont(int style, String font) {
    if (style < -1 || style >= styles.length) return;
    if (style == -1) { 
      for (int i = 0; i < styles.length; i++) {
        StyleConstants.setFontFamily(styles[i], font);
      }
    }
    else StyleConstants.setFontFamily(styles[style], font);
  }
  
    
  /**
   * Replaces the text between the <code>start</code> index and the <code>end
   * </code> index by the given string. It is important to note that 
   * <code>start</code> and <code>end</code> actually point <b>between</b>
   * caracters, therefore <code>start</code> points just before the first
   * caracter, and <code>end</code> points just after the last one.
   * @param start the start position
   * @param end the end position
   * @param newText the text that will replace the old one
   */
  public void replace(int start, int end, String newText) {
    try {
      doc.replace(start, end-start, newText, styles[defaultStyle]);
    }
    catch (BadLocationException e) {
      System.out.println(e);
    }      
  }

  
  /**
   * Modifies the scollbars if necessary in order to be sure that the character
   * specified by its position is visible.
   * @param pos the position of the character
   */  
  public void scrollToVisible(int pos) {
    try {
      tp.updateUI();
      Rectangle r = tp.modelToView(pos);
      r.y -= getHeight() / 4;
      r.height += getHeight() / 2;
      tp.scrollRectToVisible(r);
    }
    catch (BadLocationException e) {
      System.out.println(e);
    }                
  }
  
  
  /**
   * Saves the value of the vertical scollbar.
   * @see #restoreVertScrollBarPos
   */  
  public void saveVertScrollBarPos() {
    vertScrollBarPos = scrollPane.getVerticalScrollBar().getValue();
  }
  
  
  /**
   * Saves the value of the horizontal scollbar.
   * @see #restoreHorScrollBarPos
   */    
  public void saveHorScrollBarPos() {
    horScrollBarPos = scrollPane.getHorizontalScrollBar().getValue();
  }
    

  /**
   * Saves the values of the vertical and horizontal scollbars.
   * @see #restoreScrollBarPos
   */    
  public void saveScrollBarPos() {
    saveVertScrollBarPos();
    saveHorScrollBarPos(); 
  }
  
  
  /**
   * Restores the value of the vertical scollbar.
   * @see #saveVertScrollBarPos
   */    
  public void restoreVertScrollBarPos() {
    EventQueue.invokeLater(new Runnable() {
      public void run() {
        scrollPane.getVerticalScrollBar().setValue(vertScrollBarPos);
      }
    });
  }
   

  /**
   * Restores the value of the horizontal scollbar.
   * @see #saveHorScrollBarPos
   */       
  public void restoreHorScrollBarPos() {
    EventQueue.invokeLater(new Runnable() {
      public void run() {
        scrollPane.getHorizontalScrollBar().setValue(horScrollBarPos);
      }
    });
  }
  

  /**
   * Restores the values of the vertical and the horizontal scollbars.
   * @see #saveScrollBarPos
   */    
  public void restoreScrollBarPos() {
    restoreVertScrollBarPos();
    restoreHorScrollBarPos();
  }
}